import re
from django.contrib.auth.models import Group
from django.core.management.base import BaseCommand
from django.db.utils import IntegrityError
from sqlalchemy import create_engine, text
from sf_auth.models import User
from core.models.client import Company
from core.models.location import Location
from core.models.product import (
    Finish,
    StairTreadStyle,
    LandingTreadStyle,
    RailStyle,
    MiscProduct,
)
from core.models.project import (
    Project,
    ProjectCategory,
    Scope,
)


class Command(BaseCommand):
    help = 'Imports objects from legacy database'
    
    def __init__(self):
        super().__init__()
        self._queries_ran = 0

    def add_arguments(self, parser):
        parser.add_argument('--superuser', help='The email of a superuser')
        parser.add_argument(
            '--db-user',
            help='The database user',
            default='pacif244_psc',
        )
        parser.add_argument(
            '--db-pass',
            help='The database password',
            default='ded6514psc',
        )
        parser.add_argument(
            '--db-host',
            help='The database host',
            default='10.10.48.76:3306',
        )
        parser.add_argument(
            '--db-name',
            help='The database name',
            default='pacif244_psc',
        )
        parser.add_argument(
            '--first-project',
            help='The earliest project to import',
            default='1500',
        )

    def _q(self, query_string):
        self._queries_ran += 1
        with self.engine.connect() as conn:
            return conn.execute(text(query_string))

    def handle(self, *args, **options):
        user = options['db_user']
        passwd = options['db_pass']
        host = options['db_host']
        db = options['db_name']
        self.engine = create_engine(
            f"mysql+mysqlconnector://{user}:{passwd}@{host}/{db}"
        )
        self.stdout.write('Deleting builtin groups')
        Group.objects.all().delete()
        self.import_users(*args, **options)
        self.import_project_categories(*args, **options)
        # these aren't actually that useful
        # self.import_finishes()
        self.fix_scope_floor_landing_count()
        self.import_projects(*args, **options)
        self.stdout.write('''
Import summary:

%d users (%d roles)
%d projects (%d categories)
%d locations
%d companies
%d finishes

%d total queries
        ''' % (
            User.objects.count(),
            Group.objects.count(),
            Project.objects.count(),
            ProjectCategory.objects.count(),
            Location.objects.count(),
            Company.objects.count(),
            0, # Finish.objects.count(),
            self._queries_ran,
        ))

    def import_users(self, *args, **options):
        users = self._q('''
        select uid,name,mail
        from fqi_users where status=1 and login!=0
        ''')
        superuser = options['superuser']
        for uid, name, mail in users:
            name = name.split(' ')
            first_name = name[0]
            try:
                last_name = name[1]
            except IndexError:
                last_name = None

            u, new = User.objects.get_or_create(
                email=mail,
                first_name=first_name,
                last_name=last_name,
                legacy_id=uid,
            )
            u.is_admin = True
            if new:
                self.stdout.write('\nImporting new user: %s %s <%s>' % (
                    first_name,
                    last_name,
                    mail,
                ))
                if mail == superuser:
                    self.stdout.write(f'Making {mail} a superuser')
                    u.is_superuser = True
                self.import_user_roles(u)
                u.save()

    def import_user_roles(self, user):
        query = f'''
        SELECT name
        FROM fqi_role
        	JOIN fqi_users_roles ON fqi_role.rid=fqi_users_roles.rid
        where uid={user.legacy_id}
        '''
        for name, in self._q(query):
            group, new = Group.objects.get_or_create(name=name)
            if new:
                self.stdout.write(f'Importing role {name}')
            self.stdout.write(f'Adding role {name} to {user}')
            user.groups.add(group)
            user.save()

    def import_project_categories(self, *args, **options):
        query = 'SELECT name,tid FROM fqi_taxonomy_term_data WHERE vid=2'
        self.stdout.write('')
        for legacy_category in self._q(query):
            c, new = ProjectCategory.objects.get_or_create(
                name=legacy_category.name,
                legacy_id=legacy_category.tid,
            )
            if new:
                self.stdout.write(f'Importing project category: {c.name}')
                c.save()

    def set_project_categories(self, project):
        query = f'''
        SELECT field_project_project_category_tid AS tid
        FROM fqi_field_data_field_project_project_category
        WHERE entity_id={project.legacy_id}
        '''
        for category in self._q(query):
            c = ProjectCategory.objects.get(legacy_id=category.tid)
            project.categories.add(c)

    def import_projects(self, *args, **options):
        first = options['first_project']
        # fields that need no processing
        basic_fields = [
            'name',
            'bid_number',
            'contract_amount',
            'actual_amount',
            'client_job_number',
            'client_contract_no',
            'building_code',
            'notes',
        ]
        # add in fields that do need processing
        fields = basic_fields + [
            'job_number',
            'award_date',
            'complete_date',
        ]
        wheres = [
            f'field_project_job_number_value >= {first}',
        ]
        locs = ['address', 'mail_address', 'ship_address']
        project_q = query_factory('project')
        query = project_q(fields, wheres, locs)
        projects = self._q(query)
        for legacy_project in projects:
            project, new = Project.objects.get_or_create(
                number=legacy_project.job_number,
            )
            if new:
                self.stdout.write(f'\nImporting project {project.number}')
                project.legacy_id = legacy_project.nid
                project.status = self.get_project_status(project.number)
                if legacy_project.address:
                    project.address = self.import_location(
                        legacy_project.address,
                        f'Job {project.number} Site',
                    )
                if legacy_project.mail_address:
                    project.mail_address = self.import_location(
                        legacy_project.mail_address,
                        f'Job {project.number} Mailing',
                    )
                if legacy_project.ship_address:
                    project.ship_address = self.import_location(
                        legacy_project.address,
                        f'Job {project.number} Shipping',
                    )
                if legacy_project.award_date:
                    project.award_date = legacy_project.award_date.date()
                if legacy_project.complete_date:
                    project.complete_date = legacy_project.complete_date.date()
                for field in basic_fields:
                    setattr(project, field, getattr(legacy_project, field))
                self.set_project_categories(project)
                self.import_project_companies(project)
                project.project_manager = self.get_project_pm(project)
                project.save()
                self.import_project_scopes(project)

    def normalize_scope_name(self, name):
        name = name.strip()
        if re.match('^[0-9]*$', name):
            normalized_name = f'Stair {name}'
            self.stdout.write(f'Normalizing scope name: {name} -> {normalized_name}')
            return normalized_name
        return name

    def fix_scope_floor_landing_count(self):
        # no idea why this one field is in a table with a nonstandard name
        self.stdout.write('Fixing scope field floor_level_count')
        self._q('''
        CREATE OR REPLACE VIEW
        fqi_field_data_field_scope_floor_landing_count
        AS
        SELECT entity_id,
               field_floor_landing_count_value AS field_scope_floor_landing_count_value
        FROM fqi_field_data_field_floor_landing_count
        ''')

    def import_project_scopes(self, project):
        self.stdout.write(f'Importing scopes for {project.number}')
        query = f'''
        SELECT field_project_scope_target_id as tid
        FROM fqi_field_data_field_project_scope
        WHERE entity_id={project.legacy_id}
        '''
        scope_q = query_factory('scope')
        for scope_id in self._q(query):
            scope, new = Scope.objects.get_or_create(
                project=project,
                legacy_id=scope_id.tid,
            )
            if new:
                query = scope_q(
                    [
                        'stair_number',
                        'notes',
                        'flight_count',
                        'mid_landing_count',
                        'floor_landing_count',
                    ],
                    [f'nid={scope_id.tid}'],
                    [],
                )
                try:
                    legacy_scope = list(self._q(query))[0]
                except IndexError:
                    continue
                scope.name = self.normalize_scope_name(legacy_scope.stair_number)
                self.stdout.write(f'Importing scope {scope.name}')
                scope.notes = legacy_scope.notes
                scope.flight_count = legacy_scope.flight_count or 0
                scope.mid_landing_count = legacy_scope.mid_landing_count or 0
                scope.floor_landing_count = legacy_scope.floor_landing_count or 0
                # self.set_scope_finishes(scope)
                self.import_scope_products(scope)
                scope.save()

    def set_scope_finishes(self, scope):
        query = f'''
        SELECT name
        FROM fqi_field_data_field_product_category_finishes
        	JOIN fqi_taxonomy_term_data
        		ON field_product_category_finishes_tid=tid
        WHERE entity_id={scope.legacy_id}
        '''
        for name, in self._q(query):
            self.stdout.write(f'Adding finish {name} to scope {scope.name}')
            scope.finishes.add(Finish.objects.get(name=name))

    def import_scope_products(self, scope):
        query = f'''
        SELECT name, parent
        FROM fqi_taxonomy_term_data
        	JOIN fqi_field_data_field_product_category
        		ON field_product_category_tid=tid
        	JOIN fqi_taxonomy_term_hierarchy
        		on fqi_taxonomy_term_data.tid=fqi_taxonomy_term_hierarchy.tid
        WHERE entity_id={scope.legacy_id}
        AND name not in ('Stairs', 'Landings', 'Rails', 'Miscellaneous')
        '''
        for name, parent in self._q(query):
            if name == 'Specialty':
                if parent == 1:
                    ptype = StairTreadStyle
                    field = scope.stair_styles
                elif parent == 3:
                    ptype = LandingTreadStyle
                    field = scope.land_styles
                elif parent == 4:
                    ptype = MiscProduct
                    field = scope.rail_styles
            elif re.match(r'^S[0-9]00', name):
                ptype = StairTreadStyle
                field = scope.stair_styles
            elif re.match(r'^L[0-9]00', name):
                ptype = LandingTreadStyle
                field = scope.land_styles
            elif re.match(r'^R[0-9]00', name):
                ptype = RailStyle
                field = scope.rail_styles
            else:
                ptype = MiscProduct
                field = scope.misc_prods
            try:
                prod = ptype.objects.get(name=name)
            except ptype.DoesNotExist:
                self.stdout.write(f'Ignoring unknown product: {name}')
                continue
            self.stdout.write(f'Adding product {prod.name} to scope {scope.name}')
            field.add(prod)

    def get_project_status(self, number):
        query = '''
        SELECT fqi_taxonomy_term_data.name
        FROM fqi_node
    		JOIN fqi_field_data_field_project_status
    			ON fqi_field_data_field_project_status.entity_id = fqi_node.nid
    		JOIN fqi_taxonomy_term_data
    			ON fqi_taxonomy_term_data.tid =
        		fqi_field_data_field_project_status.field_project_status_tid
        '''
        return self._q(query).first().name

    def import_project_companies(self, project):
        for c in [
                'owner',
                'lender',
                'general_contractor',
                'architect',
                'engineer',
                'client',
        ]:
            query = f'''
            SELECT title AS name
            FROM fqi_field_data_field_project_{c}
            	JOIN fqi_node ON field_project_{c}_target_id=nid
            WHERE entity_id={project.legacy_id}
            '''
            try:
                company_name = list(self._q(query))[0].name
            except IndexError:
                continue
            setattr(project, c, self.import_company(company_name))
            self.stdout.write(f'Setting {company_name} as {c} for {project.number}')

    def import_company(self, name):
        company, new = Company.objects.get_or_create(name=name)
        if new:
            self.stdout.write(f'Importing company {company.name}')
        return company

    def import_location(self, lid, default_name=''):
        query = f'''
        SELECT name,street,city,province,postal_code,additional AS notes
        FROM fqi_location
        	JOIN fqi_location_email
        WHERE fqi_location.lid={lid}
        '''
        loc, new = Location.objects.get_or_create(legacy_id=lid)
        if new:
            legacy_loc = self._q(query).first()
            if legacy_loc.name:
                loc.name = legacy_loc.name
            else:
                loc.name = default_name
            loc.street = legacy_loc.street
            loc.city = legacy_loc.city
            loc.province = legacy_loc.province[:2]
            loc.postal_code = legacy_loc.postal_code
            self.stdout.write(f'Attempting import of location {loc}')
            try:
                loc.save()
                self.stdout.write('Imported')
            except IntegrityError:
                self.stdout.write('Apparent duplicate found -- aborting')
                loc.delete()
                return Location.objects.get(
                    street=loc.street,
                    city=loc.city,
                    province=loc.province,
                )
        return loc

    def get_project_pm(self, project):
        query = f'''
        SELECT field_project_project_manager_target_id AS id
        FROM fqi_field_data_field_project_project_manager
        WHERE entity_id={project.legacy_id}
        '''
        try:
            return User.objects.get(legacy_id=list(self._q(query))[0].id)
        except User.DoesNotExist:
            return None

    def import_finishes(self):
        query = 'SELECT name FROM fqi_taxonomy_term_data WHERE vid=22'
        self.stdout.write('')
        for name, in self._q(query):
            self.stdout.write(f'Importing finish: {name}')
            Finish.objects.create(name=name)


def query_factory(node):
    def qfunc(fields, wheres, locs):
        field_columns = [f'field_{node}_{f}_value AS {f}' for f in fields]
        loc_columns = [f'field_{node}_{loc}_lid AS {loc}' for loc in locs]
        columns = ",\n".join(field_columns + loc_columns)
        prefix = f'fqi_field_data_field_{node}'
        joins = '\n'.join([
            f'LEFT JOIN {prefix}_{f} ON nid={prefix}_{f}.entity_id'
            for f in (fields + locs)
        ])
        wheres = '\n'.join(wheres)
        return f'''
SELECT nid,
{columns}

FROM fqi_node
{joins}

WHERE fqi_node.type="{node}"
AND
{wheres}
'''
    return qfunc
