from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.db import models
from django.forms.widgets import Select, CheckboxSelectMultiple
from django.utils import timezone
from functools import reduce
from modelcluster.models import ClusterableModel
from modelcluster.fields import ParentalKey
from wagtail.admin.panels import FieldPanel, FieldRowPanel
from wagtail.documents.models import AbstractDocument
from wagtail.models import Orderable
from wagtail.search import index
from sf_auth.models import User
from .client import Company
from .product import LandingPostType
from .location import STATES


class SalesRep(models.Model):
    name = models.CharField(max_length=255)
    user = models.ForeignKey(User, blank=True, null=True, on_delete=models.SET_NULL)

    def __str__(self):
        return self.name


class ScopeSummaryMixin:
    def summary(self):
        summaries = [scope.summary() for scope in self.scopes.all()]
        return reduce(BaseBidScope.sum_dicts, summaries, {})


no_space_validator = RegexValidator(
    r' ',
    'No spaces allowed',
    inverse_match=True,
    code='invalid_bid_number',
)


class Bid(ScopeSummaryMixin, index.Indexed, ClusterableModel):
    number = models.CharField(
        max_length=16,
        unique=True,
        help_text='not including revision, e.g. "23-100"',
        validators=[no_space_validator],
    )
    name = models.CharField(max_length=255)
    estimator = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        limit_choices_to={'groups__name__in': ['staff estimator']},
    )
    sales_rep = models.ForeignKey(SalesRep, on_delete=models.PROTECT)
    city = models.CharField(max_length=255, blank=True, null=True)
    province = models.CharField(
        'state',
        max_length=255,
        blank=True,
        null=True,
        choices=STATES,
    )
    client = models.ForeignKey(
        Company,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='client_bids',
    )
    contractor = models.ForeignKey(
        Company,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='contractor_bids',
    )
    architect = models.ForeignKey(
        Company,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='architect_bids',
    )
    engineer = models.ForeignKey(
        Company,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='engineer_bids',
    )
    installed = models.BooleanField(default=False)

    @property
    def current_revision(self):
        return self.revisions.count()

    def make_new_revision(self):
        rev = BidRevision.objects.create(
            bid=self,
            number=self.current_revision,
        )
        self.save()
        for scope in self.scopes.all():
            RevScope.objects.create(
                revision=rev,
                name=scope.name,
                dwg_ref=scope.dwg_ref,
                grid=scope.grid,
                slab_construction=scope.slab_construction,
                takeoff=scope.takeoff,
            )
        for doc in self.documents.all():
            RevDocument.objects.create(
                revision=rev,
                file=doc.file,
                title=doc.title,
                issue_date=doc.issue_date,
                received_date=doc.received_date,
                type=doc.type,
            )

    def __str__(self):
        return f'{self.number} {self.name}'


class BidRevision(ScopeSummaryMixin, Orderable, ClusterableModel):
    class Meta:
        ordering = ['-number']

    bid = ParentalKey(Bid, on_delete=models.CASCADE, related_name='revisions')
    number = models.IntegerField()
    date = models.DateField(default=timezone.now)

    def __str__(self):
        return f'{self.bid.number} R{self.number}'

    def name(self):
        return f'Rev {self.number}'


class BidRevChildMixin:
    def get_bid(self):
        try:
            return self.bid
        except AttributeError:
            return self.revision.bid


class BaseBidScope(BidRevChildMixin, Orderable, ClusterableModel):
    class Meta:
        abstract = True

    name = models.CharField(max_length=255)
    dwg_ref = models.CharField(
        'drawing reference',
        max_length=255,
        blank=True,
        null=True,
    )
    grid = models.CharField(max_length=16, blank=True, null=True)
    slab_construction = models.ForeignKey('core.SlabConstruction', on_delete=models.PROTECT)
    takeoff = models.JSONField(null=True)

    TAKEOFF_KEYS = {
        'risers': 'Risers',
        'flights': 'Flights',
        'ex_treads': 'Extended Treads',
        'haunches': 'Haunches',
        'mid_landings': 'Mid Landings',
        'floor_landings': 'Floor Landings',
        'columns': 'Columns',
        'cane_rail': 'Cane Rail',
        'guard_rail': 'Guard Rail',
        'center_rail': 'Center Rail',
        'ex_sr': 'Extra Stair Rail',
        'ex_wr': 'Extra Wall Rail',
        'add_sr': 'Added Stair Rail',
        'add_wr': 'Added Wall Rail',
        'gates': 'Gates',
        'standpipes': 'Standpipes',
    }

    TAKEOFF_SCHEMA = {
        'type': 'array',
        'items': {
            'type': 'object',
            'description': 'A bid scope takeoff',
            'properties':
            {'name': {'type': 'string'}} | {k: {'type': 'number'} for k in TAKEOFF_KEYS}
        },
    }

    panels = [
        FieldRowPanel([
            FieldPanel('name'),
            FieldPanel('slab_construction', widget=Select),
        ]),
        FieldRowPanel([
            FieldPanel('grid'),
            FieldPanel('dwg_ref'),
        ]),
    ]

    def sum_dicts(d1, d2):
        return {
            k: d1.get(k, 0) + d2.get(k, 0)
            for k in (set(d1) | set(d2)) - set(['name'])
        }

    def summary(self):
        if self.takeoff:
            return reduce(BaseBidScope.sum_dicts, self.takeoff, {})
        return {}


class BidScope(BaseBidScope):
    bid = ParentalKey(Bid, related_name='scopes')

    def __str__(self):
        return f'{self.bid} {self.name}'


class RevScope(BaseBidScope):
    revision = ParentalKey(BidRevision, related_name='scopes')

    def __str__(self):
        return f'{self.revision} {self.name}'


def _get_bid_upload_path(instance, filename):
    bid = instance.get_bid()
    rev = bid.current_revision
    return f'bids/{bid.number}/rev{rev}/{filename}'


class BaseBidDocument(BidRevChildMixin, Orderable, ClusterableModel):
    class Meta:
        abstract = True
        verbose_name = 'bid document'

    file = models.FileField(upload_to=_get_bid_upload_path)
    title = models.CharField(max_length=255)
    issue_date = models.DateField()
    received_date = models.DateField(default=timezone.now)

    TYPE_CHOICES = [
        ('drawing', 'Drawing'),
        ('addendum', 'Addendum'),
        ('specs', 'Specs'),
        ('other', 'Other'),
    ]
    type = models.CharField(verbose_name='type', max_length=32, choices=TYPE_CHOICES)

    panels = [
        FieldPanel('file'),
        FieldRowPanel([
            FieldPanel('title'),
            FieldPanel('type'),
        ]),
        FieldRowPanel([
            FieldPanel('issue_date'),
            FieldPanel('received_date'),
        ]),
    ]

    def name(self):
        return self.title

    def count_file_uses(file):
        b = BidDocument.objects.filter(file=file).count()
        r = RevDocument.objects.filter(file=file).count()
        return b + r


class BidDocument(BaseBidDocument):
    bid = ParentalKey(Bid, on_delete=models.CASCADE, related_name='documents')


class RevDocument(BaseBidDocument):
    revision = ParentalKey(BidRevision, on_delete=models.CASCADE, related_name='documents')


# class BaseAdder(Orderable, ClusterableModel):
#     TYPE_CHOICES = [
#         ('each', 'Each'),
#         ('floor', 'Per floor'),
#         ('flight', 'Per flight'),
#         ('riser', 'Per riser'),
#     ]
