from django.db import models
from modelcluster.models import ClusterableModel
from modelcluster.fields import ParentalKey
from wagtail.admin.panels import ObjectList, MultiFieldPanel, FieldPanel, FieldRowPanel
from wagtail.fields import RichTextField
from wagtail.models import Orderable
from .util import ChoiceList, cash_field


# special properties only for galv finish
class GalvMixin(models.Model):
    class Meta:
        abstract = True

    galv_base_price = cash_field(
        verbose_name='base price (stairs/landings)',
        help_text='per pound',
        blank=True,
        null=True,
    )
    galv_base_price_rail = cash_field(
        verbose_name='base price (rails)',
        help_text='per pound',
        blank=True,
        null=True,
    )
    galv_base_price_column = cash_field(
        verbose_name='base price (columns)',
        help_text='per pound',
        blank=True,
        null=True,
    )
    galv_pounds_riser = cash_field(
        verbose_name='pounds per riser-foot',
        blank=True,
        null=True,
    )
    galv_pounds_sqft = cash_field(
        verbose_name='pounds per landing sqft',
        blank=True,
        null=True,
    )
    galv_pounds_sr = cash_field(
        verbose_name='pounds per linear foot',
        help_text='stair rail and guard rail',
        blank=True,
        null=True,
    )
    galv_pounds_wr = cash_field(
        verbose_name='pounds per linear foot',
        help_text='wall rail',
        blank=True,
        null=True,
    )
    galv_pounds_column = cash_field(
        verbose_name='pounds per linear foot',
        help_text='columns',
        blank=True,
        null=True,
    )

    galv_panel = ObjectList(
        [
            MultiFieldPanel([
                FieldRowPanel([
                    FieldPanel('galv_base_price'),
                    FieldPanel('galv_base_price_rail'),
                    FieldPanel('galv_base_price_column'),
                ]),
                FieldRowPanel([
                    FieldPanel('galv_pounds_riser'),
                    FieldPanel('galv_pounds_sqft'),
                ]),
                FieldRowPanel([
                    FieldPanel('galv_pounds_sr'),
                    FieldPanel('galv_pounds_wr'),
                    FieldPanel('galv_pounds_column'),
                ]),
            ]),
        ],
        heading='Galv Pricing',
    )


class Finish(GalvMixin, ChoiceList):
    class Meta(ChoiceList.Meta):
        verbose_name_plural = 'finishes'

    price_riser = cash_field(verbose_name='price per riser', blank=True, null=True)
    price_sqft = cash_field(verbose_name='price per landing sqft', blank=True, null=True)
    price_ex_sr = cash_field(
        verbose_name='extra stair rail price',
        help_text='per linear foot',
        blank=True,
        null=True,
    )
    price_ex_wr = cash_field(
        verbose_name='extra wall rail price',
        help_text='per linear foot',
        blank=True,
        null=True,
    )


class StringerType(ChoiceList):
    pass


class LandingPostType(ChoiceList):
    description = models.CharField(max_length=255)
    weight = models.DecimalField(
        help_text='pounds per foot',
        max_digits=6,
        decimal_places=1,
    )


class Product(ChoiceList):
    description = RichTextField(blank=True, null=True)

    class Meta(ChoiceList.Meta):
        abstract = True


class StairTreadStyle(Product):
    pass


class StairPricePoint(models.Model):
    tread_style = models.ForeignKey(StairTreadStyle, on_delete=models.CASCADE)
    stringer_type = models.ForeignKey(StringerType, on_delete=models.CASCADE)
    width = models.IntegerField(help_text='inches')
    price = cash_field(help_text='per riser')


class LandingTreadStyle(Product):
    base_price = cash_field(help_text='per square foot', blank=True, null=True)
    floor_price_adj = cash_field(
        verbose_name='floor-level adjustment',
        help_text='relative to mid-level landings, per square foot',
        blank=True,
        null=True,
    )
    stair_style = models.ForeignKey(
        StairTreadStyle,
        verbose_name='corresponding stair tread style',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    def floor_price(self):
        if self.floor_price_adj:
            return self.base_price + self.floor_price_adj
        return self.base_price

    def clean(self):
        self.description = self.stair_style.description


class RailStyle(Product):
    base_price = cash_field(help_text='per linear foot', blank=True, null=True)
    add_price_adj = cash_field(
        verbose_name='add price adjust',
        help_text='difference from base price for added/CIP, per linear foot',
        blank=True,
        null=True,
    )

    def add_price(self):
        if self.add_price_adj:
            return self.base_price + self.add_price_adj
        return self.base_price


class MiscProduct(Product):
    pass


