from django.db import models
from django.forms.widgets import CheckboxSelectMultiple
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from modelcluster.models import ClusterableModel
from wagtail.admin.panels import (
    FieldPanel,
    FieldRowPanel,
    MultiFieldPanel,
    TabbedInterface,
    ObjectList,
)
from wagtail.contrib.routable_page.models import RoutablePageMixin, path
from wagtail.fields import RichTextField
from wagtail.models import Page, Orderable
from wagtail.search import index
from .client import Company
from .frontend import LandingPage
from .location import Location
from .product import (
    LandingTreadStyle,
    Finish,
    MiscProduct,
    RailStyle,
    StairTreadStyle,
)
from .util import ChoiceList, LegacyModel, cash_field
from sf_auth.models import User


class BuildingCode(ChoiceList):
    name = models.CharField(max_length=255)
    notes = RichTextField()


class SlabConstruction(ChoiceList):
    name = models.CharField(max_length=255)
    install_price = cash_field(help_text='per floor', blank=True, null=True)
    uses_embeds = models.BooleanField()
    uses_hangers = models.BooleanField()


class ProjectCategory(LegacyModel, ChoiceList):
    class Meta(ChoiceList.Meta):
        verbose_name_plural = 'project categories'


class Project(LegacyModel, index.Indexed, ClusterableModel):
    number = models.IntegerField(unique=True)
    name = models.CharField(max_length=255)
    categories = models.ManyToManyField(ProjectCategory)
    STATUS_CHOICES = [
        ('Awarded', 'Awarded'),
        ('Ongoing', 'Ongoing'),
        ('Complete', 'Complete'),
        ('On Hold', 'On Hold'),
        ('Canceled', 'Canceled'),
    ]
    status = models.CharField(
        max_length=255,
        choices=STATUS_CHOICES,
        null=True,
    )
    award_date = models.DateField(null=True, blank=True)
    complete_date = models.DateField(null=True, blank=True)
    bid_number = models.CharField(max_length=255, blank=True, null=True)
    contract_amount = cash_field(blank=True, null=True)
    actual_amount = cash_field(verbose_name='closeout amount', blank=True, null=True)

    client_job_number = models.CharField(max_length=255, blank=True, null=True)
    client_contract_no = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name='client contract number',
    )
    building_code = models.CharField(max_length=255, blank=True, null=True)
    notes = models.TextField(blank=True, null=True)

    address = models.ForeignKey(
        Location,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    mail_address = models.ForeignKey(
        Location,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='mail_address'
    )
    ship_address = models.ForeignKey(
        Location,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='ship_address'
    )

    owner = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='owner_projects'
    )

    lender = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='lender_projects'
    )

    general_contractor = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='contractor_projects'
    )

    architect = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='architect_projects'
    )

    engineer = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='engineer_projects',
        verbose_name='engineer of record',
    )

    client = models.ForeignKey(
        Company,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='client_projects'
    )

    project_manager = models.ForeignKey(
        User,
        limit_choices_to={'groups__name__in': ['staff project manager']},
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    search_fields = [
        index.SearchField('name'),
        index.AutocompleteField('name'),
        index.SearchField('number'),
        index.AutocompleteField('number'),
        index.FilterField('status'),
    ]

    def __str__(self):
        return f'{self.number} - {self.name}'

    class Meta:
        ordering = ['-number']


class ProjectIndexPage(RoutablePageMixin, Page):
    show_in_menus_default = True
    parent_page_types = [LandingPage]
    subpage_types = []

    @path('')
    def index(self, request):
        return self.render(request, context_overrides={
            'projects': Project.objects.all(),
        })

    @path('<int:number>/')
    def by_number(self, request, number):
        return self.render(
            request,
            context_overrides={
                'project': Project.objects.get(number=number),
            },
            template='core/project_page.html',
        )


class Scope(LegacyModel, index.Indexed, Orderable, ClusterableModel):
    class Meta:
        ordering = ['project__number', 'sort_order', 'name']

    project = ParentalKey(Project, on_delete=models.CASCADE, related_name='scopes')
    name = models.CharField(
        max_length=255,
        help_text='e.g. "Stair 1" or "Machine Room Stair"'
    )
    alt_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text='Stair name in client documents, if different',
    )
    grid = models.CharField(max_length=16, blank=True, null=True)
    levels = models.CharField(max_length=64, blank=True, null=True)
    notes = models.TextField(blank=True, null=True)

    finishes = ParentalManyToManyField(Finish, blank=True)
    stair_styles = ParentalManyToManyField(
        StairTreadStyle,
        blank=True,
        verbose_name='stair tread styles',
    )
    land_styles = ParentalManyToManyField(
        LandingTreadStyle,
        blank=True,
        verbose_name='landing tread styles',
    )
    rail_styles = ParentalManyToManyField(
        RailStyle,
        blank=True,
        verbose_name='rail styles',
    )
    misc_prods = ParentalManyToManyField(
        MiscProduct,
        blank=True,
        verbose_name='misc products',
    )

    flight_count = models.IntegerField(default=0)
    mid_landing_count = models.IntegerField(default=0)
    floor_landing_count = models.IntegerField(default=0)

    def landing_count(self):
        return self.mid_landing_count + self.floor_landing_count

    def __str__(self):
        return f'{self.project.number} {self.name}'

    panels = [
        FieldRowPanel([
            FieldPanel('name'),
            FieldPanel('alt_name'),
        ]),
        FieldRowPanel([
            FieldPanel('grid'),
            FieldPanel('levels'),
        ]),
        FieldPanel('notes'),
        FieldRowPanel([
            FieldPanel('flight_count'),
            FieldPanel('mid_landing_count'),
            FieldPanel('floor_landing_count'),
        ]),
        MultiFieldPanel([
            FieldRowPanel([
                FieldPanel('stair_styles', widget=CheckboxSelectMultiple),
                FieldPanel('land_styles', widget=CheckboxSelectMultiple),
                FieldPanel('rail_styles', widget=CheckboxSelectMultiple),
                FieldPanel('misc_prods', widget=CheckboxSelectMultiple),
            ]),
            FieldRowPanel([
                FieldPanel('finishes', widget=CheckboxSelectMultiple),
            ]),
        ]),
    ]

    search_fields = [
        index.SearchField('name'),
        index.AutocompleteField('name'),
        index.SearchField('notes'),
        index.AutocompleteField('notes'),

        index.FilterField('stair_styles'),
        index.FilterField('land_styles'),
        index.FilterField('rail_styles'),
        index.FilterField('misc_prods'),

        index.RelatedFields('project', [
            index.SearchField('name'),
            index.AutocompleteField('name'),
            index.SearchField('number'),
            index.AutocompleteField('number'),
        ]),
    ]

