import json
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from django.db.models import Count
from django.forms.widgets import Select, CheckboxSelectMultiple
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from wagtail.admin.panels import (
    Panel,
    FieldPanel,
    FieldRowPanel,
    MultiFieldPanel,
    TabbedInterface,
    ObjectList,
    InlinePanel,
)
from wagtail.admin.views.reports import ReportView
from wagtail.admin.viewsets.chooser import ChooserViewSet
from wagtail.snippets.views.snippets import (
    SnippetViewSet,
    SnippetViewSetGroup,
    CreateView,
    EditView,
)
from .panels import TemplatePanel, CurrentRevision, ReadOnlyInlinePanel
from .models.bid import (
    SalesRep,
    Bid,
    BidDocument,
    BidRevision,
    BidScope,
)
from .models.client import Company
from .models.location import Location
from .models.product import (
    StairTreadStyle,
    LandingTreadStyle,
    RailStyle,
    MiscProduct,
    Finish,
)
from .models.project import (
    Project,
    ProjectCategory,
    SlabConstruction,
    Scope,
)


class SelfRedirectEditView(EditView):
    def get_success_url(self):
        return self.get_edit_url()


SnippetViewSet.list_per_page = 50


class ProjectViewSet(SnippetViewSet):
    model = Project
    icon = 'helmet'
    list_display = ['__str__', 'status', 'project_manager']
    edit_view_class = SelfRedirectEditView

    edit_handler = TabbedInterface([
        ObjectList(
            [
                MultiFieldPanel([
                    FieldRowPanel([
                        FieldPanel('number'),
                        FieldPanel('name'),
                    ]),
                    FieldRowPanel([
                        FieldPanel('project_manager'),
                    ]),
                ]),
                MultiFieldPanel(
                    heading='Progress',
                    children=[
                        FieldPanel('status'),
                        FieldRowPanel(children=[
                            FieldPanel('bid_number'),
                            FieldPanel('award_date'),
                            FieldPanel('complete_date'),
                        ]),
                    ]),
                MultiFieldPanel(
                    heading='Misc',
                    children=[
                        FieldPanel('contract_amount'),
                        FieldPanel('actual_amount'),
                        FieldPanel('building_code'),
                        FieldPanel('notes'),
                    ]),
                FieldPanel('categories', widget=CheckboxSelectMultiple),
            ],
            heading="Basic Info",
        ),
        ObjectList(
            [
                MultiFieldPanel(
                    children=[
                        FieldPanel('address'),
                        FieldPanel('mail_address'),
                        FieldPanel('ship_address'),
                    ]),
            ],
            heading='Addresses'
        ),
        ObjectList(
            [
                MultiFieldPanel(
                    children=[
                        FieldPanel('client'),
                        FieldRowPanel(children=[
                            FieldPanel('client_job_number'),
                            FieldPanel('client_contract_no'),
                        ]),
                        FieldPanel('owner'),
                        FieldPanel('lender'),
                        FieldPanel('general_contractor'),
                        FieldPanel('architect'),
                        FieldPanel('engineer'),
                    ]),
            ],
            heading='Relationships'
        ),
        ObjectList(
            [
                InlinePanel('scopes'),
            ],
            heading='Scope',
        ),
    ])


class ScopeViewSet(SnippetViewSet):
    model = Scope
    icon = 'crosshairs'
    list_display = ['name', 'project']
    list_filter = ['stair_styles', 'land_styles', 'rail_styles', 'misc_prods']


class ProjectCategoryViewSet(SnippetViewSet):
    model = ProjectCategory
    icon = 'form'
    list_display = ['name']

    panels = [
        FieldPanel('name'),
        FieldPanel('sort_order'),
    ]


class LocationChooserViewSet(ChooserViewSet):
    model = Location
    form_fields = [
        'name',
        'street',
        'city',
        'province',
        'postal_code',
        'notes',
    ]


class LocationViewSet(SnippetViewSet):
    model = Location
    icon = 'location'
    chooser_viewset_class = LocationChooserViewSet
    list_display = ['name', 'street', 'city', 'province', 'postal_code']
    list_filter = ['city', 'postal_code', 'province']
    search_fields = ['name', 'street']

    panels = [
        FieldPanel('name'),
        FieldPanel('street'),
        FieldPanel('city'),
        FieldPanel('province'),
        FieldPanel('postal_code'),
        FieldPanel('notes'),
    ]


class CompanyChooserViewSet(ChooserViewSet):
    model = Company
    form_fields = ['name']


class CompanyViewSet(SnippetViewSet):
    model = Company
    icon = 'building'
    chooser_viewset_class = CompanyChooserViewSet
    list_display = ['name']

    panels = [
        FieldPanel('name'),
    ]


class ProjectViewSetGroup(SnippetViewSetGroup):
    menu_label = 'Projects'
    add_to_admin_menu = True
    items = [
        ProjectViewSet,
        ScopeViewSet,
        ProjectCategoryViewSet,
        LocationViewSet,
        CompanyViewSet,
    ]


class ProductViewSet(SnippetViewSet):
    list_display = ['name']
    frontend_panels_tab = ObjectList(
        [
            FieldPanel('name'),
            FieldPanel('description'),
            FieldPanel('sort_order'),
        ],
        heading='Frontend',
    )


class StairTreadStyleViewSet(ProductViewSet):
    model = StairTreadStyle
    edit_handler = TabbedInterface([ProductViewSet.frontend_panels_tab])


class LandingTreadStyleViewSet(ProductViewSet):
    model = LandingTreadStyle
    list_display = ['name', 'base_price', 'floor_price']
    edit_handler = TabbedInterface([
        ObjectList(
            [
                FieldPanel('name'),
                FieldPanel('stair_style'),
                FieldPanel('sort_order'),
            ],
            heading='Frontend',
        ),
        ObjectList(
            [
                MultiFieldPanel([
                    FieldPanel('base_price'),
                    FieldPanel('floor_price_adj'),
                ]),
            ],
            heading='Pricing',
        )
    ])


class RailStyleViewSet(ProductViewSet):
    model = RailStyle
    list_display = ['name', 'base_price', 'add_price']
    edit_handler = TabbedInterface([
        ProductViewSet.frontend_panels_tab,
        ObjectList(
            [
                MultiFieldPanel([
                    FieldPanel('base_price'),
                    FieldPanel('add_price_adj'),
                ])
            ],
            heading='Pricing',
        )
    ])


class MiscProductViewSet(ProductViewSet):
    model = MiscProduct


class FinishViewSet(SnippetViewSet):
    model = Finish
    create_template_name = 'core/finish_edit.html'
    edit_template_name = 'core/finish_edit.html'

    def get_edit_handler(self):
        panels = TabbedInterface([
            MultiFieldPanel(
                [
                    FieldPanel('name'),
                    FieldPanel('sort_order'),
                ],
                heading='Basic Info',
            ),
            MultiFieldPanel(
                [
                    FieldPanel('price_riser'),
                    FieldPanel('price_sqft'),
                    FieldPanel('price_ex_sr'),
                    FieldPanel('price_ex_wr'),
                ],
                heading='Pricing',
            ),
            Finish.galv_panel,
        ])
        return panels.bind_to_model(self.model)


class SlabConstructionViewSet(SnippetViewSet):
    model = SlabConstruction
    panels = [
        MultiFieldPanel([
            FieldPanel('name'),
            FieldPanel('install_price'),
            FieldPanel('uses_embeds'),
            FieldPanel('uses_hangers'),
            FieldPanel('sort_order'),
        ]),
    ]


class ProductViewSetGroup(SnippetViewSetGroup):
    items = [
        StairTreadStyleViewSet,
        LandingTreadStyleViewSet,
        RailStyleViewSet,
        MiscProductViewSet,
        FinishViewSet,
        SlabConstructionViewSet,
    ]
    menu_label = 'Products'
    add_to_admin_menu = True


class BidEditView(EditView):
    def get_success_url(self):
        return self.get_edit_url() + '#tab-current_revision'

    def post(self, request, *args, **kwargs):
        response = super().post(request, *args, **kwargs)
        if self.get_form().is_valid():
            bid = self.get_object()
            if 'create-bid-rev' in request.POST:
                bid.make_new_revision()
            elif 'delete-last-bid-rev' in request.POST:
                bid.revisions.first().delete()
        return response


class BidViewSet(SnippetViewSet):
    model = Bid
    icon = 'openquote'
    edit_view_class = BidEditView
    edit_handler = TabbedInterface([
        ObjectList(
            [
                MultiFieldPanel([
                    FieldPanel('number'),
                    FieldPanel('name'),
                    FieldRowPanel([
                        FieldPanel('city'),
                        FieldPanel('province'),
                    ]),
                ]),
                FieldRowPanel([
                    FieldPanel('estimator'),
                    FieldPanel('sales_rep', widget=Select),
                ]),
                FieldRowPanel([
                    FieldPanel('client'),
                    FieldPanel('contractor'),
                ]),
                FieldRowPanel([
                    FieldPanel('architect'),
                    FieldPanel('engineer'),
                ]),
            ],
            heading='Basic info',
        ),
        CurrentRevision(
            [
                TemplatePanel('core/takeoff_summary.html'),
                InlinePanel(
                    'documents',
                    heading='Supporting Documents',
                    classname='collapsed',
                ),
                InlinePanel(
                    'scopes',
                    panels=BidScope.panels + [
                        TemplatePanel('core/takeoff_edit_link.html'),
                        TemplatePanel('core/takeoff_summary.html'),
                    ],
                    heading='Scopes',
                ),
            ],
        ),
        ObjectList(
            [
                ReadOnlyInlinePanel(
                    'revisions',
                    panels=[
                        MultiFieldPanel([
                            FieldPanel('date', read_only=True),
                            TemplatePanel('core/takeoff_summary.html'),
                        ]),
                        ReadOnlyInlinePanel(
                            'documents',
                            heading='Supporting Documents',
                        ),
                        ReadOnlyInlinePanel(
                            'scopes',
                            heading='Scopes',
                            panels = BidScope.panels + [
                                TemplatePanel('core/takeoff_summary.html'),
                            ],
                        ),
                    ],
                ),
            ],
            heading='Past Revisions',
        ),
    ])


class SalesRepViewSet(SnippetViewSet):
    model = SalesRep
    icon = 'group'


class BidViewSetGroup(SnippetViewSetGroup):
    menu_label = 'Estimating'
    add_to_admin_menu = True
    items = [
        BidViewSet,
        SalesRepViewSet,
    ]


@csrf_exempt
def edit_takeoff(request, pk):
    scope = BidScope.objects.get(pk=pk)
    if request.method == 'POST':
        try:
            takeoff = json.load(request)
            validate(takeoff, BidScope.TAKEOFF_SCHEMA)
        except ValidationError as e:
            response = HttpResponseBadRequest()
            response.reason_phrase = str(e)
            return response
        scope.takeoff = takeoff
        scope.save()
        return HttpResponse()
    return render(request, 'core/takeoff_edit.html', context={'scope': scope})
